local lmath     = require "lib.lmath"
local wavefront = require "lib.wavefront"
local fps       = require "lib.fps"
require "myDpad"

-------------------------------------------------------------------------------

local math_min  = math.min
local math_max  = math.max
local math_abs  = math.abs
local math_rad  = math.rad
local math_sqrt = math.sqrt

-------------------------------------------------------------------------------

local temp_mat4  = lmath.matrix4.new()
local temp_mat4b = lmath.matrix4.new()

-------------------------------------------------------------------------------

local shaders={}

shaders.simple=love.graphics.newShader(
	"shaders/simple_vertex.glsl",
	"shaders/simple_pixel.glsl"
)

-------------------------------------------------------------------------------

local mesh_format={}

mesh_format.static={
	{"VertexPosition","float",3},
	{"VertexNormal","float",3},
	{"VertexTexture","float",2}
}

-------------------------------------------------------------------------------

local meshes={}

local function load_mesh(data)
	local mesh={}
	
	mesh.data=wavefront.import(data)
	
	local min_x,min_y,min_z = 0,0,0
	local max_x,max_y,max_z = 0,0,0
	
	--Calculate boundary
	for i=1,#mesh.data.vertices,3 do
		local vx = mesh.data.vertices[i]
		local vy = mesh.data.vertices[i+1]
		local vz = mesh.data.vertices[i+2]
		
		min_x = math_min(vx,min_x)
		min_y = math_min(vy,min_y)
		min_z = math_min(vz,min_z)
		
		max_x = math_max(vx,max_x)
		max_y = math_max(vy,max_y)
		max_z = math_max(vz,max_z)
	end
	
	local size_x = max_x-min_x
	local size_y = max_y-min_y
	local size_z = max_z-min_z
	
	local middle_x = (min_x+max_x)/2
	local middle_y = (min_y+max_y)/2
	local middle_z = (min_z+max_z)/2
	
	do
		local vertices={}
		
		local vertices_={}
		
		for i=1,#mesh.data.vertices,3 do
			local vx = mesh.data.vertices[i]
			local vy = mesh.data.vertices[i+1]
			local vz = mesh.data.vertices[i+2]
			
			vertices_[i]   = (vx-middle_x)/size_x
			vertices_[i+1] = (vy-middle_y)/size_y
			vertices_[i+2] = (vz-middle_z)/size_z
		end
		
		for i=1,#mesh.data.faces,3 do
			vertices[#vertices+1]={
				vertices_[(mesh.data.faces[i]-1)*3+1],
				vertices_[(mesh.data.faces[i]-1)*3+2],
				vertices_[(mesh.data.faces[i]-1)*3+3],
				mesh.data.normals[(mesh.data.faces[i+1]-1)*3+1],
				mesh.data.normals[(mesh.data.faces[i+1]-1)*3+2],
				mesh.data.normals[(mesh.data.faces[i+1]-1)*3+3],
				mesh.data.textures[(mesh.data.faces[i+2]-1)*2+1],
				mesh.data.textures[(mesh.data.faces[i+2]-1)*2+2]
			}
		end
		
		mesh.drawable=love.graphics.newMesh(
			mesh_format.static,vertices,
			"triangles","static"
		)
	end
	
	do		
		local faces={}
		
		for i=1,#mesh.data.faces,3 do
			faces[#faces+1]=mesh.data.faces[i]
		end

		mesh.shape=fps.shape.new(mesh.data.vertices,faces)
	end
	
	return mesh
end

-------------------------------------------------------------------------------

meshes.pyramid = load_mesh(love.filesystem.read("assets/meshes/pyramid.obj"))
meshes.cube    = load_mesh(love.filesystem.read("assets/meshes/cube.obj"))
meshes.cone    = load_mesh(love.filesystem.read("assets/meshes/cone.obj"))
meshes.sphere  = load_mesh(love.filesystem.read("assets/meshes/sphere.obj"))

-------------------------------------------------------------------------------

local projection=lmath.matrix4.new():set_perspective(
	75,
	love.graphics.getWidth()/love.graphics.getHeight(),
	0.1,1000
)

local camera=lmath.matrix4.new():set_position(0,5,0)

local camera_step
local camera_input

do
	local key_map={
		up=lmath.vector3.new(0,0,-1),
		down=lmath.vector3.new(0,0,1),
		left=lmath.vector3.new(-1,0,0),
		right=lmath.vector3.new(1,0,0),
		b=lmath.vector3.new(0,1,0),
		a=lmath.vector3.new(0,-1,0)
	}
	
	local move_dir=lmath.vector3.new()
	local look_angle=lmath.vector2.new()
	
	local view=lmath.matrix4.new()
	:set(camera:unpack())
	
	camera_step=function(dt)
		move_dir:set()
		for key,vector in pairs(key_map) do
			if DPAD.isDown[key]then
				move_dir:add(vector)
			end
		end
		move_dir=move_dir
		:normalize()
		:multiply(dt*(love.keyboard.isDown("lshift") and 40 or 20))
		if move_dir:get_magnitude()~=0 then
			view:transform(move_dir:unpack())
		end
		camera:lerp(view,dt/0.1)
	end
	
	camera_input=function(dx,dy)
		if love.mouse.isDown(1,2) then
			look_angle.x=lmath.clamp(
				look_angle.x-math.rad(dy*0.4),
				math.rad(-89),math.rad(89)
			)
			look_angle.y=look_angle.y-math.rad(dx*0.4)
			view:set_euler(0,look_angle.y,0)
			:rotate_euler(look_angle.x,0,0)
		end
	end
end

-------------------------------------------------------------------------------

local buffer={
	love.graphics.newCanvas(
		love.graphics.getWidth(),
		love.graphics.getHeight(),
		{format="normal"}
	),
	depth=true
}

-------------------------------------------------------------------------------

local wireframe    = false
local bounding_box = false
local step_physics = false

local objects={}

local world=fps.world.new() --Setup new physics scene
world:add_solver(fps.solvers.rigid)
world:set_gravity(0,-50,0)

local function new_object(fixtures,transform,color)
	local object={
		fixtures = {},
		body     = fps.body.new(),
		color    = color or lmath.color3.new(1,1,1),
	}
	
	object.body:set_transform(transform:unpack())
	
	for _,fixture in pairs(fixtures) do
		local collider=fps.collider.new()
		collider:set_shape(fixture[1].shape)
		collider:set_size(fixture[2]:unpack())
		collider:set_transform(fixture[3]:unpack())
		object.body:add_collider(collider)
		
		object.fixtures[collider]=fixture[1]
	end
	
	world:add_body(object.body)
	
	objects[object.body]=object
	
	return object
end

local function draw_objects()
	temp_mat4:set(camera:unpack()):inverse()
	
	love.graphics.push("all")
	love.graphics.setCanvas(buffer)
	love.graphics.clear(0,0,0,0)
	love.graphics.setDepthMode("lequal",true)
	love.graphics.setShader(shaders.simple)
	shaders.simple:send("projection","row",projection)
	shaders.simple:send("view","row",temp_mat4)
	
	for _,object in pairs(objects) do
		if wireframe then
			love.graphics.setWireframe(true)
			love.graphics.setMeshCullMode("none")
		else
			love.graphics.setWireframe(false)
			love.graphics.setMeshCullMode("front")
		end
		
		local body=object.body
		local boundary=body.boundary
		
		for _,collider in ipairs(body.colliders) do
			temp_mat4
			:set(unpack(body.transform))
			:multiply(collider.transform)
			:scale(
				collider.size[1],
				collider.size[2],
				collider.size[3]
			)
			
			shaders.simple:send("model","row",temp_mat4)
			
			love.graphics.setColor(object.color.r,object.color.g,object.color.b,1)
			
			love.graphics.draw(object.fixtures[collider].drawable)
		end
		
		local middle_x = (boundary[1]+boundary[4])/2
		local middle_y = (boundary[2]+boundary[5])/2
		local middle_z = (boundary[3]+boundary[6])/2
		
		temp_mat4
		:set()
		:set_position(middle_x,middle_y,middle_z)
		:scale(
			boundary[4]-boundary[1],
			boundary[5]-boundary[2],
			boundary[6]-boundary[3]
		)
		
		if bounding_box then
			love.graphics.setWireframe(true)
			love.graphics.setMeshCullMode("none")
			shaders.simple:send("model","row",temp_mat4)
			love.graphics.setColor(0,1,0,1)
			love.graphics.draw(meshes.cube.drawable)
		end
	end
	
	love.graphics.pop()
	love.graphics.draw(
		buffer[1],
		0,buffer[1]:getHeight(),
		0,1,-1
	)
end

-------------------------------------------------------------------------------

new_object( --Setup the base platform
	{
		{
			meshes.cube,
			lmath.vector3.new(30,1,30),
			lmath.matrix4.new()
		}
	},
	lmath.matrix4.new(),
	lmath.color3.new(1,1,1)
)

new_object(
	{
		{
			meshes.cube,
			lmath.vector3.new(30,10,1),
			lmath.matrix4.new()
		}
	},
	lmath.matrix4.new():set_position(0,5,-15),
	lmath.color3.new(1,1,1)
)

new_object(
	{
		{
			meshes.cube,
			lmath.vector3.new(30,10,1),
			lmath.matrix4.new()
		}
	},
	lmath.matrix4.new():set_position(0,5,15),
	lmath.color3.new(1,1,1)
)

new_object(
	{
		{
			meshes.cube,
			lmath.vector3.new(1,10,30),
			lmath.matrix4.new()
		}
	},
	lmath.matrix4.new():set_position(-15,5,0),
	lmath.color3.new(1,1,1)
)

new_object(
	{
		{
			meshes.cube,
			lmath.vector3.new(1,10,30),
			lmath.matrix4.new()
		}
	},
	lmath.matrix4.new():set_position(15,5,0),
	lmath.color3.new(1,1,1)
)

-------------------------------------------------------------------------------

function love.draw(dt)
	draw_objects()
	
	love.graphics.print(
		(
			"FPS: %d\n"..
			"Mem: %0.2f mb\n"..
			"Bodies: %d\n"..
			"Wireframe: %s\n"..
			"Show Bounding Box: %s\n"..
			"Physics: %s"
		):format(
			love.timer.getFPS(),
			collectgarbage("count")/1024,
			#world.bodies,
			tostring(wireframe),
			tostring(bounding_box),
			step_physics and "Step" or "Realtime"
		),
		5,5
	)
	
	DPAD:draw()
end

local tick=love.timer.getTime()
local scale=1

local pull_origin = lmath.matrix4.new()
local pull_force  = 50

function love.update(dt)
  DPAD:update(dt)
  
	camera_step(dt)
	
	pull_origin
	:set(camera:unpack())
	:transform(0,0,-15)
	
	local pull_x,pull_y,pull_z = pull_origin:get_position()
	
	if DPAD.isDown['y'] then
		for _,object in pairs(objects) do
			if not object.body:get_static() then
				local x,y,z = object.body:get_position()
				
				local mass = object.body:get_mass()
				
				object.body:apply_force(
					(pull_x-x)*mass*pull_force,
					(pull_y-y)*mass*pull_force,
					(pull_z-z)*mass*pull_force
				)
				
				--Dampen the velocity so they stop swooshing around
				object.body.velocity[1]=object.body.velocity[1]*0.99
				object.body.velocity[2]=object.body.velocity[2]*0.99
				object.body.velocity[3]=object.body.velocity[3]*0.99
				
				object.body.angular_velocity[1]=object.body.angular_velocity[1]*0.99
				object.body.angular_velocity[2]=object.body.angular_velocity[2]*0.99
				object.body.angular_velocity[3]=object.body.angular_velocity[3]*0.99
			end
		end
	end
	
	if not step_physics then
		world:step(math.min(dt,1/60))
	end
end

function love.mousemoved(x,y,dx,dy)
	camera_input(dx,dy)
end

function love.keypressed(key)
	if key=="f" then
		local lx,ly,lz=camera:get_look()
		
		local mesh=math.random(1,4)
		
		local bullet=new_object(
			{
				{
					(mesh==1 and meshes.cube) or
					(mesh==2 and meshes.pyramid) or
					(mesh==3 and meshes.cone) or
					(mesh==4 and meshes.sphere),
					lmath.vector3.new(3,3,3),
					lmath.matrix4.new()
				}
			},
			lmath.matrix4.new()
			:set(camera:unpack())
			:transform(0,0,-6),
			lmath.color3.new(
				math.random(0,255)/255,
				math.random(0,255)/255,
				math.random(0,255)/255
			)
		)
		
		bullet.body:set_static(false)
		bullet.body:set_velocity(
			lx*20,ly*20,lz*20
		)
	elseif key=="1" then
		wireframe=not wireframe
	elseif key=="2" then
		bounding_box=not bounding_box
	elseif key=="3" then
		step_physics=not step_physics
	elseif key=="t" and step_physics then
		world:step(1/60)
	end
	
end
function love.mousereleased( x, y, button, istouch, presses )
  if istouch then
    if DPAD.isDown['x'] then

		local lx,ly,lz=camera:get_look()

		
		local mesh=math.random(1,4)
		
		local bullet=new_object(
			{
				{
					(mesh==1 and meshes.cube) or
					(mesh==2 and meshes.pyramid) or
					(mesh==3 and meshes.cone) or
					(mesh==4 and meshes.sphere),
					lmath.vector3.new(3,3,3),
					lmath.matrix4.new()
				}
			},
			lmath.matrix4.new()
			:set(camera:unpack())
			:transform(0,0,-6),
			lmath.color3.new(
				math.random(0,255)/255,
				math.random(0,255)/255,
				math.random(0,255)/255
			)
		)
		
		bullet.body:set_static(false)
		bullet.body:set_velocity(
			lx*20,ly*20,lz*20
		)
	end
  end
end

collectgarbage()